-
Notifications
You must be signed in to change notification settings - Fork 232
Expand file tree
/
Copy pathfp16_fwd.cpp
More file actions
179 lines (146 loc) · 6.82 KB
/
Copy pathfp16_fwd.cpp
File metadata and controls
179 lines (146 loc) · 6.82 KB
1
2
3
4
5
6
7
8
9
10
11
12
13
14
15
16
17
18
19
20
21
22
23
24
25
26
27
28
29
30
31
32
33
34
35
36
37
38
39
40
41
42
43
44
45
46
47
48
49
50
51
52
53
54
55
56
57
58
59
60
61
62
63
64
65
66
67
68
69
70
71
72
73
74
75
76
77
78
79
80
81
82
83
84
85
86
87
88
89
90
91
92
93
94
95
96
97
98
99
100
101
102
103
104
105
106
107
108
109
110
111
112
113
114
115
116
117
118
119
120
121
122
123
124
125
126
127
128
129
130
131
132
133
134
135
136
137
138
139
140
141
142
143
144
145
146
147
148
149
150
151
152
153
154
155
156
157
158
159
160
161
162
163
164
165
166
167
168
169
170
171
172
173
174
175
176
177
178
179
/*
* SPDX-FileCopyrightText: Copyright (c) 2023 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
* SPDX-License-Identifier: MIT
*/
#include <catch2/catch_test_macros.hpp>
#include "../utils/helpers.h"
#include <cuda_runtime_api.h>
#include <cudnn_frontend.h>
namespace fe = cudnn_frontend;
/*
Run this example by using command:
bin/samples "Toy sdpa forward"
This example shows how to construct a sdpa forward graph.
*/
// Tensors in forward pass
#define Q_UID 1
#define K_UID 2
#define V_UID 3
#define O_UID 4
#define STATS_UID 5
#define BIAS_UID 6
#define SEQ_LEN_Q_UID 7
#define SEQ_LEN_KV_UID 8
std::shared_ptr<fe::graph::Graph>
create_sdpa_forward_graph(int64_t const b,
int64_t const h_q,
int64_t const h_k,
int64_t const h_v,
int64_t const s_q,
int64_t const s_kv,
int64_t const d_qk,
int64_t const d_v,
float const attn_scale = 1.0f,
bool const generate_stats = true,
bool const causal_mask = false,
bool const padding_mask = false) {
// Create a graph and set common global properties.
auto graph = std::make_shared<fe::graph::Graph>();
graph->set_io_data_type(fe::DataType_t::BFLOAT16)
.set_intermediate_data_type(fe::DataType_t::FLOAT)
.set_compute_data_type(fe::DataType_t::FLOAT);
auto Q = graph->tensor(fe::graph::Tensor_attributes()
.set_name("Q")
.set_uid(Q_UID)
.set_dim({b, h_q, s_q, d_qk})
.set_stride({h_q * s_q * d_qk, s_q * d_qk, d_qk, 1}));
auto K = graph->tensor(fe::graph::Tensor_attributes()
.set_name("K")
.set_uid(K_UID)
.set_dim({b, h_k, s_kv, d_qk})
.set_stride({h_k * s_kv * d_qk, s_kv * d_qk, d_qk, 1}));
auto V = graph->tensor(fe::graph::Tensor_attributes()
.set_name("V")
.set_uid(V_UID)
.set_dim({b, h_v, s_kv, d_v})
.set_stride({h_v * s_kv * d_v, s_kv * d_v, d_v, 1}));
auto sdpa_options = fe::graph::SDPA_attributes()
.set_name("flash_attention")
.set_generate_stats(generate_stats)
.set_attn_scale(attn_scale);
if (causal_mask) {
sdpa_options.set_diagonal_alignment(cudnn_frontend::DiagonalAlignment_t::TOP_LEFT)
.set_diagonal_band_right_bound(0);
}
if (padding_mask) {
auto seq_q = graph->tensor(fe::graph::Tensor_attributes()
.set_name("seq_q")
.set_uid(SEQ_LEN_Q_UID)
.set_dim({b, 1, 1, 1})
.set_stride({1, 1, 1, 1})
.set_data_type(fe::DataType_t::INT32));
auto seq_kv = graph->tensor(fe::graph::Tensor_attributes()
.set_name("seq_kv")
.set_uid(SEQ_LEN_KV_UID)
.set_dim({b, 1, 1, 1})
.set_stride({1, 1, 1, 1})
.set_data_type(fe::DataType_t::INT32));
sdpa_options.set_padding_mask(padding_mask).set_seq_len_q(seq_q).set_seq_len_kv(seq_kv);
}
auto [O, Stats] = graph->sdpa(Q, K, V, sdpa_options);
O->set_output(true).set_dim({b, h_q, s_q, d_v}).set_stride({h_q * d_v, d_v, b * h_q * d_v, 1}).set_uid(O_UID);
if (generate_stats) {
Stats->set_output(true).set_data_type(fe::DataType_t::FLOAT).set_uid(STATS_UID);
} else {
assert(Stats == nullptr);
}
return graph;
}
TEST_CASE("Toy sdpa forward", "[graph][sdpa][flash][forward]") {
int64_t b = 3; // batch size
int64_t h_q = 4; // head dim
int64_t h_k = 4; // head dim
int64_t h_v = 4; // head dim
int64_t s_q = 1024; // q tensor is padded to this seq length
int64_t s_kv = 1024; // k and v tensor is padded to this seq length
int64_t d_qk = 128; // hidden dim
int64_t d_v = 128; // hidden dim
bool generate_stats = true;
float attn_scale = 0.123f;
bool causal_mask = true;
bool padding_mask = (cudnnGetVersion() >= 8903);
if (cudnnGetVersion() < 8903) {
SKIP("Test requires cudnn 8.9.3 or above");
return;
}
// Create a unique_ptr for the cuDNN handle
auto handle_ptr = create_cudnn_handle();
auto handle = *handle_ptr;
auto graph = create_sdpa_forward_graph(
b, h_q, h_k, h_v, s_q, s_kv, d_qk, d_v, attn_scale, generate_stats, causal_mask, padding_mask);
REQUIRE(graph->build(handle, {fe::HeurMode_t::A}).is_good());
//// Build variant pack
Surface<half> q_tensor(b * h_q * s_q * d_qk);
Surface<half> k_tensor(b * h_k * d_qk * s_kv);
Surface<half> v_tensor(b * h_v * d_v * s_kv);
Surface<half> o_tensor(b * s_q * h_q * d_qk);
std::unordered_map<fe::graph::Tensor_attributes::uid_t, void*> variant_pack = {
{Q_UID, q_tensor.devPtr}, {K_UID, k_tensor.devPtr}, {V_UID, v_tensor.devPtr}, {O_UID, o_tensor.devPtr}};
Surface<half> bias_tensor(b * 1 * s_q * s_kv);
Surface<int32_t> devActualSeqlenQ(b);
Surface<int32_t> devActualSeqlenKV(b);
if (padding_mask) {
std::vector<int32_t> hostActualSeqlenQ(b, 20);
std::vector<int32_t> hostActualSeqlenKV(b, 20);
CUDA_CHECK(cudaMemcpy(devActualSeqlenQ.devPtr,
hostActualSeqlenQ.data(),
sizeof(hostActualSeqlenQ[0]) * b,
cudaMemcpyHostToDevice));
CUDA_CHECK(cudaMemcpy(devActualSeqlenKV.devPtr,
hostActualSeqlenKV.data(),
sizeof(hostActualSeqlenKV[0]) * b,
cudaMemcpyHostToDevice));
CUDA_CHECK(cudaDeviceSynchronize());
variant_pack[SEQ_LEN_Q_UID] = devActualSeqlenQ.devPtr;
variant_pack[SEQ_LEN_KV_UID] = devActualSeqlenKV.devPtr;
}
Surface<float> statsTensor(b * h_q * s_q * 1);
if (generate_stats == true) {
variant_pack[STATS_UID] = statsTensor.devPtr;
}
int64_t workspace_size = 0;
REQUIRE(graph->get_workspace_size(workspace_size).is_good());
Surface<int8_t> workspace(workspace_size);
REQUIRE(graph->execute(handle, variant_pack, workspace.devPtr).is_good());
CUDA_CHECK(cudaDeviceSynchronize());
}